@@ -2,6 +2,7 @@ from __future__ import print_function, division, absolute_import
 from __future__ import unicode_literals
 from fontTools.feaLib.error import FeatureLibError
 from fontTools.feaLib.lexer import Lexer, IncludingLexer
+from fontTools.misc.encodingTools import getEncoding
 from fontTools.misc.py23 import *
 import fontTools.feaLib.ast as ast
 import logging
@@ -13,11 +14,23 @@ log = logging.getLogger(__name__)
 
 
 class Parser(object):
-
     extensions = {}
     ast = ast
 
-    def __init__(self, featurefile):
+    def __init__(self, featurefile, glyphNames=(), **kwargs):
+        if "glyphMap" in kwargs:
+            from fontTools.misc.loggingTools import deprecateArgument
+            deprecateArgument("glyphMap", "use 'glyphNames' (iterable) instead")
+            if glyphNames:
+                raise TypeError("'glyphNames' and (deprecated) 'glyphMap' are "
+                                "mutually exclusive")
+            glyphNames = kwargs.pop("glyphMap")
+        if kwargs:
+            raise TypeError("unsupported keyword argument%s: %s"
+                            % ("" if len(kwargs) == 1 else "s",
+                               ", ".join(repr(k) for k in kwargs)))
+
+        self.glyphNames_ = set(glyphNames)
         self.doc_ = self.ast.FeatureFile()
         self.anchors_ = SymbolTable()
         self.glyphclasses_ = SymbolTable()
@@ -27,15 +40,18 @@ class Parser(object):
             self.anchors_, self.valuerecords_
         }
         self.next_token_type_, self.next_token_ = (None, None)
+        self.cur_comments_ = []
         self.next_token_location_ = None
         self.lexer_ = IncludingLexer(featurefile)
-        self.advance_lexer_()
+        self.advance_lexer_(comments=True)
 
     def parse(self):
         statements = self.doc_.statements
         while self.next_token_type_ is not None:
-            self.advance_lexer_()
-            if self.cur_token_type_ is Lexer.GLYPHCLASS:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.cur_token_type_ is Lexer.GLYPHCLASS:
                 statements.append(self.parse_glyphclass_definition_())
             elif self.is_cur_keyword_(("anon", "anonymous")):
                 statements.append(self.parse_anonymous_())
@@ -61,7 +77,7 @@ class Parser(object):
             else:
                 raise FeatureLibError(
                     "Expected feature, languagesystem, lookup, markClass, "
-                    "table, or glyph class definition",
+                    "table, or glyph class definition, got {} \"{}\"".format(self.cur_token_type_, self.cur_token_),
                     self.cur_token_location_)
         return self.doc_
 
@@ -193,6 +209,45 @@ class Parser(object):
         self.glyphclasses_.define(name, glyphclass)
         return glyphclass
 
+    def split_glyph_range_(self, name, location):
+        # Since v1.20, the OpenType Feature File specification allows
+        # for dashes in glyph names. A sequence like "a-b-c-d" could
+        # therefore mean a single glyph whose name happens to be
+        # "a-b-c-d", or it could mean a range from glyph "a" to glyph
+        # "b-c-d", or a range from glyph "a-b" to glyph "c-d", or a
+        # range from glyph "a-b-c" to glyph "d".Technically, this
+        # example could be resolved because the (pretty complex)
+        # definition of glyph ranges renders most of these splits
+        # invalid. But the specification does not say that a compiler
+        # should try to apply such fancy heuristics. To encourage
+        # unambiguous feature files, we therefore try all possible
+        # splits and reject the feature file if there are multiple
+        # splits possible. It is intentional that we don't just emit a
+        # warning; warnings tend to get ignored. To fix the problem,
+        # font designers can trivially add spaces around the intended
+        # split point, and we emit a compiler error that suggests
+        # how exactly the source should be rewritten to make things
+        # unambiguous.
+        parts = name.split("-")
+        solutions = []
+        for i in range(len(parts)):
+            start, limit = "-".join(parts[0:i]), "-".join(parts[i:])
+            if start in self.glyphNames_ and limit in self.glyphNames_:
+                solutions.append((start, limit))
+        if len(solutions) == 1:
+            start, limit = solutions[0]
+            return start, limit
+        elif len(solutions) == 0:
+            raise FeatureLibError(
+                "\"%s\" is not a glyph in the font, and it can not be split "
+                "into a range of known glyphs" % name, location)
+        else:
+            ranges = " or ".join(["\"%s - %s\"" % (s, l) for s, l in solutions])
+            raise FeatureLibError(
+                "Ambiguous glyph range \"%s\"; "
+                "please use %s to clarify what you mean" % (name, ranges),
+                location)
+
     def parse_glyphclass_(self, accept_glyphname):
         if (accept_glyphname and
                 self.next_token_type_ in (Lexer.NAME, Lexer.CID)):
@@ -216,14 +271,19 @@ class Parser(object):
         while self.next_token_ != "]":
             if self.next_token_type_ is Lexer.NAME:
                 glyph = self.expect_glyph_()
-                if self.next_token_ == "-":
-                    range_location = self.cur_token_location_
-                    range_start = glyph
+                location = self.cur_token_location_
+                if '-' in glyph and glyph not in self.glyphNames_:
+                    start, limit = self.split_glyph_range_(glyph, location)
+                    glyphs.add_range(
+                        start, limit,
+                        self.make_glyph_range_(location, start, limit))
+                elif self.next_token_ == "-":
+                    start = glyph
                     self.expect_symbol_("-")
-                    range_end = self.expect_glyph_()
-                    glyphs.add_range(range_start, range_end,
-                                     self.make_glyph_range_(range_location,
-                                                            range_start, range_end))
+                    limit = self.expect_glyph_()
+                    glyphs.add_range(
+                        start, limit,
+                        self.make_glyph_range_(location, start, limit))
                 else:
                     glyphs.append(glyph)
             elif self.next_token_type_ is Lexer.CID:
@@ -245,6 +305,10 @@ class Parser(object):
                     raise FeatureLibError(
                         "Unknown glyph class @%s" % self.cur_token_,
                         self.cur_token_location_)
+                if isinstance(gc, self.ast.MarkClass):
+                    gc = self.ast.MarkClassName(self.cur_token_location_, gc)
+                else:
+                    gc = self.ast.GlyphClassName(self.cur_token_location_, gc)
                 glyphs.add_class(gc)
             else:
                 raise FeatureLibError(
@@ -276,6 +340,14 @@ class Parser(object):
                 self.expect_symbol_("'")
                 hasMarks = marked = True
             if marked:
+                if suffix:
+                    # makeotf also reports this as an error, while FontForge
+                    # silently inserts ' in all the intervening glyphs.
+                    # https://github.com/fonttools/fonttools/pull/1096
+                    raise FeatureLibError(
+                        "Unsupported contextual target sequence: at most "
+                        "one run of marked (') glyph/class names allowed",
+                        self.cur_token_location_)
                 glyphs.append(gc)
             elif glyphs:
                 suffix.append(gc)
@@ -741,9 +813,11 @@ class Parser(object):
 
     def parse_table_GDEF_(self, table):
         statements = table.statements
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.is_cur_keyword_("Attach"):
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.is_cur_keyword_("Attach"):
                 statements.append(self.parse_attach_())
             elif self.is_cur_keyword_("GlyphClassDef"):
                 statements.append(self.parse_GlyphClassDef_())
@@ -751,6 +825,8 @@ class Parser(object):
                 statements.append(self.parse_ligatureCaretByIndex_())
             elif self.is_cur_keyword_("LigatureCaretByPos"):
                 statements.append(self.parse_ligatureCaretByPos_())
+            elif self.cur_token_ == ";":
+                continue
             else:
                 raise FeatureLibError(
                     "Expected Attach, LigatureCaretByIndex, "
@@ -759,10 +835,14 @@ class Parser(object):
 
     def parse_table_head_(self, table):
         statements = table.statements
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.is_cur_keyword_("FontRevision"):
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.is_cur_keyword_("FontRevision"):
                 statements.append(self.parse_FontRevision_())
+            elif self.cur_token_ == ";":
+                continue
             else:
                 raise FeatureLibError("Expected FontRevision",
                                       self.cur_token_location_)
@@ -770,13 +850,17 @@ class Parser(object):
     def parse_table_hhea_(self, table):
         statements = table.statements
         fields = ("CaretOffset", "Ascender", "Descender", "LineGap")
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.cur_token_type_ is Lexer.NAME and self.cur_token_ in fields:
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.cur_token_type_ is Lexer.NAME and self.cur_token_ in fields:
                 key = self.cur_token_.lower()
                 value = self.expect_number_()
                 statements.append(
                     self.ast.HheaField(self.cur_token_location_, key, value))
+                if self.next_token_ != ";":
+                    raise FeatureLibError("Incomplete statement", self.next_token_location_)
             elif self.cur_token_ == ";":
                 continue
             else:
@@ -787,13 +871,17 @@ class Parser(object):
     def parse_table_vhea_(self, table):
         statements = table.statements
         fields = ("VertTypoAscender", "VertTypoDescender", "VertTypoLineGap")
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.cur_token_type_ is Lexer.NAME and self.cur_token_ in fields:
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.cur_token_type_ is Lexer.NAME and self.cur_token_ in fields:
                 key = self.cur_token_.lower()
                 value = self.expect_number_()
                 statements.append(
                     self.ast.VheaField(self.cur_token_location_, key, value))
+                if self.next_token_ != ";":
+                    raise FeatureLibError("Incomplete statement", self.next_token_location_)
             elif self.cur_token_ == ";":
                 continue
             else:
@@ -803,12 +891,16 @@ class Parser(object):
 
     def parse_table_name_(self, table):
         statements = table.statements
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.is_cur_keyword_("nameid"):
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.is_cur_keyword_("nameid"):
                 statement = self.parse_nameid_()
                 if statement:
                     statements.append(statement)
+            elif self.cur_token_ == ";":
+                continue
             else:
                 raise FeatureLibError("Expected nameid",
                                       self.cur_token_location_)
@@ -818,14 +910,15 @@ class Parser(object):
         langID = None
         if self.next_token_type_ == Lexer.NUMBER:
             platformID = self.expect_number_()
+            location = self.cur_token_location_
             if platformID not in (1, 3):
-                raise FeatureLibError("Expected platform id 1 or 3",
-                                      self.cur_token_location_)
+                raise FeatureLibError("Expected platform id 1 or 3", location)
             if self.next_token_type_ == Lexer.NUMBER:
                 platEncID = self.expect_number_()
                 langID = self.expect_number_()
         else:
             platformID = 3
+            location = self.cur_token_location_
 
         if platformID == 1:                # Macintosh
             platEncID = platEncID or 0     # Roman
@@ -837,18 +930,17 @@ class Parser(object):
         string = self.expect_string_()
         self.expect_symbol_(";")
 
-        if platformID == 1 and platEncID == 0:
-            string = self.unescape_mac_name_string(string)
-        elif platformID == 3 and platEncID == 1:
-            string = self.unescape_windows_name_string(string)
-
-        return platformID, platEncID, langID, string
+        encoding = getEncoding(platformID, platEncID, langID)
+        if encoding is None:
+            raise FeatureLibError("Unsupported encoding", location)
+        unescaped = self.unescape_string_(string, encoding)
+        return platformID, platEncID, langID, unescaped
 
     def parse_nameid_(self):
         assert self.cur_token_ == "nameid", self.cur_token_
         location, nameID = self.cur_token_location_, self.expect_number_()
-        if nameID > 255:
-            raise FeatureLibError("Expected name id < 255",
+        if nameID > 32767:
+            raise FeatureLibError("Name id value cannot be greater than 32767",
                                   self.cur_token_location_)
         if 1 <= nameID <= 6:
             log.warning("Name id %d cannot be set from the feature file. "
@@ -860,27 +952,35 @@ class Parser(object):
         return self.ast.NameRecord(location, nameID, platformID, platEncID,
                                    langID, string)
 
-    def unescape_mac_name_string(self, string):
-        def unescape(match):
-            n = match.group(0)[1:]
-            c = bytechr(int(n, 16)).decode('mac_roman')
-            return c
+    def unescape_string_(self, string, encoding):
+        if encoding == "utf_16_be":
+            s = re.sub(r"\\[0-9a-fA-F]{4}", self.unescape_unichr_, string)
+        else:
+            unescape = lambda m: self.unescape_byte_(m, encoding)
+            s = re.sub(r"\\[0-9a-fA-F]{2}", unescape, string)
+        # We now have a Unicode string, but it might contain surrogate pairs.
+        # We convert surrogates to actual Unicode by round-tripping through
+        # Python's UTF-16 codec in a special mode.
+        utf16 = tobytes(s, "utf_16_be", "surrogatepass")
+        return tounicode(utf16, "utf_16_be")
 
-        return re.sub(r'\\[0-9a-fA-F]{2}', unescape, string)
+    @staticmethod
+    def unescape_unichr_(match):
+        n = match.group(0)[1:]
+        return unichr(int(n, 16))
 
-    def unescape_windows_name_string(self, string):
-        def unescape(match):
-            n = match.group(0)[1:]
-            c = unichr(int(n, 16))
-            return c
-
-        return re.sub(r'\\[0-9a-fA-F]{4}', unescape, string)
+    @staticmethod
+    def unescape_byte_(match, encoding):
+        n = match.group(0)[1:]
+        return bytechr(int(n, 16)).decode(encoding)
 
     def parse_table_BASE_(self, table):
         statements = table.statements
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.is_cur_keyword_("HorizAxis.BaseTagList"):
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.is_cur_keyword_("HorizAxis.BaseTagList"):
                 horiz_bases = self.parse_base_tag_list_()
             elif self.is_cur_keyword_("HorizAxis.BaseScriptList"):
                 horiz_scripts = self.parse_base_script_list_(len(horiz_bases))
@@ -903,9 +1003,11 @@ class Parser(object):
                    "winAscent", "winDescent", "XHeight", "CapHeight",
                    "WeightClass", "WidthClass", "LowerOpSize", "UpperOpSize")
         ranges = ("UnicodeRange", "CodePageRange")
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.cur_token_type_ is Lexer.NAME:
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.cur_token_type_ is Lexer.NAME:
                 key = self.cur_token_.lower()
                 value = None
                 if self.cur_token_ in numbers:
@@ -972,10 +1074,12 @@ class Parser(object):
         if self.next_token_type_ is Lexer.NUMBER:
             number, location = self.expect_number_(), self.cur_token_location_
             if vertical:
-                val = self.ast.ValueRecord(location, vertical, 0, 0, 0, number,
+                val = self.ast.ValueRecord(location, vertical,
+                                           None, None, None, number,
                                            None, None, None, None)
             else:
-                val = self.ast.ValueRecord(location, vertical, 0, 0, number, 0,
+                val = self.ast.ValueRecord(location, vertical,
+                                           None, None, number, None,
                                            None, None, None, None)
             return val
         self.expect_symbol_("<")
@@ -1072,27 +1176,30 @@ class Parser(object):
 
     def parse_featureNames_(self, tag):
         assert self.cur_token_ == "featureNames", self.cur_token_
+        block = self.ast.FeatureNamesBlock(self.cur_token_location_)
         self.expect_symbol_("{")
         for symtab in self.symbol_tables_:
             symtab.enter_scope()
-
-        statements = []
-        while self.next_token_ != "}":
-            self.expect_keyword_("name")
-            location = self.cur_token_location_
-            platformID, platEncID, langID, string = self.parse_name_()
-            statements.append(
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                block.statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.is_cur_keyword_("name"):
+                location = self.cur_token_location_
+                platformID, platEncID, langID, string = self.parse_name_()
+                block.statements.append(
                     self.ast.FeatureNameStatement(location, tag, platformID,
                                                   platEncID, langID, string))
-
+            elif self.cur_token_ == ";":
+                continue
+            else:
+                raise FeatureLibError('Expected "name"',
+                                      self.cur_token_location_)
         self.expect_symbol_("}")
-
         for symtab in self.symbol_tables_:
             symtab.exit_scope()
-
         self.expect_symbol_(";")
-
-        return statements
+        return block
 
     def parse_FontRevision_(self):
         assert self.cur_token_ == "FontRevision", self.cur_token_
@@ -1110,9 +1217,11 @@ class Parser(object):
             symtab.enter_scope()
 
         statements = block.statements
-        while self.next_token_ != "}":
-            self.advance_lexer_()
-            if self.cur_token_type_ is Lexer.GLYPHCLASS:
+        while self.next_token_ != "}" or self.cur_comments_:
+            self.advance_lexer_(comments=True)
+            if self.cur_token_type_ is Lexer.COMMENT:
+                statements.append(self.ast.Comment(self.cur_token_location_, self.cur_token_))
+            elif self.cur_token_type_ is Lexer.GLYPHCLASS:
                 statements.append(self.parse_glyphclass_definition_())
             elif self.is_cur_keyword_("anchorDef"):
                 statements.append(self.parse_anchordef_())
@@ -1143,7 +1252,7 @@ class Parser(object):
             elif self.is_cur_keyword_("valueRecordDef"):
                 statements.append(self.parse_valuerecord_definition_(vertical))
             elif stylisticset and self.is_cur_keyword_("featureNames"):
-                statements.extend(self.parse_featureNames_(stylisticset))
+                statements.append(self.parse_featureNames_(stylisticset))
             elif size_feature and self.is_cur_keyword_("parameters"):
                 statements.append(self.parse_size_parameters_())
             elif size_feature and self.is_cur_keyword_("sizemenuname"):
@@ -1154,7 +1263,7 @@ class Parser(object):
                 continue
             else:
                 raise FeatureLibError(
-                    "Expected glyph class definition or statement",
+                    "Expected glyph class definition or statement: got {} {}".format(self.cur_token_type_, self.cur_token_),
                     self.cur_token_location_)
 
         self.expect_symbol_("}")
@@ -1167,6 +1276,28 @@ class Parser(object):
                                   self.cur_token_location_)
         self.expect_symbol_(";")
 
+        # A multiple substitution may have a single destination, in which case
+        # it will look just like a single substitution. So if there are both
+        # multiple and single substitutions, upgrade all the single ones to
+        # multiple substitutions.
+
+        # Check if we have a mix of non-contextual singles and multiples.
+        has_single = False
+        has_multiple = False
+        for s in statements:
+            if isinstance(s, self.ast.SingleSubstStatement):
+                has_single = not any([s.prefix, s.suffix, s.forceChain])
+            elif isinstance(s, self.ast.MultipleSubstStatement):
+                has_multiple = not any([s.prefix, s.suffix])
+
+        # Upgrade all single substitutions to multiple substitutions.
+        if has_single and has_multiple:
+            for i, s in enumerate(statements):
+                if isinstance(s, self.ast.SingleSubstStatement):
+                    statements[i] = self.ast.MultipleSubstStatement(s.location,
+                        s.prefix, s.glyphs[0].glyphSet()[0], s.suffix,
+                        [r.glyphSet()[0] for r in s.replacements])
+
     def is_cur_keyword_(self, k):
         if self.cur_token_type_ is Lexer.NAME:
             if isinstance(k, type("")):  # basestring is gone in Python3
@@ -1285,14 +1416,24 @@ class Parser(object):
             return self.cur_token_
         raise FeatureLibError("Expected a string", self.cur_token_location_)
 
-    def advance_lexer_(self):
-        self.cur_token_type_, self.cur_token_, self.cur_token_location_ = (
-            self.next_token_type_, self.next_token_, self.next_token_location_)
-        try:
-            (self.next_token_type_, self.next_token_,
-             self.next_token_location_) = self.lexer_.next()
-        except StopIteration:
-            self.next_token_type_, self.next_token_ = (None, None)
+    def advance_lexer_(self, comments=False):
+        if comments and self.cur_comments_:
+            self.cur_token_type_ = Lexer.COMMENT
+            self.cur_token_, self.cur_token_location_ = self.cur_comments_.pop(0)
+            return
+        else:
+            self.cur_token_type_, self.cur_token_, self.cur_token_location_ = (
+                self.next_token_type_, self.next_token_, self.next_token_location_)
+            self.cur_comments_ = []
+        while True:
+            try:
+                (self.next_token_type_, self.next_token_,
+                 self.next_token_location_) = next(self.lexer_)
+            except StopIteration:
+                self.next_token_type_, self.next_token_ = (None, None)
+            if self.next_token_type_ != Lexer.COMMENT:
+                break
+            self.cur_comments_.append((self.next_token_, self.next_token_location_))
 
     @staticmethod
     def reverse_string_(s):
